#
# Copyright 2013 Infinidat Ltd. All rights reserved.
# Use is subject to license terms.
#


from .client import ZeroRPCClientTransport
from .server import ZeroRPCServerTransport

__all__ = ['ZeroRPCClientTransport', 'ZeroRPCServerTransport']


##
# Monkey patching zerorpc to fix two problems:
# - gevent error printing to stderr
# - Event repr bloated to an enormous size
def monkey_patch_zerorpc():
    def zerorpc_event_repr(self):
        try:
            # the __str__ method of zerorpc.Event is actually more useful than __repr__ as it does not override it
            return str(self)
        except:
            object.__repr__(self)

    # zerorpc creates instances of Event so deep, there's no better way than this
    import zerorpc
    zerorpc.Event.__repr__ = zerorpc_event_repr
    zerorpc.gevent_zmq.logger.error = lambda *args, **kwargs: None  # it prints to stderr two known issues

monkey_patch_zerorpc()



#
# Copyright 2013 Infinidat Ltd. All rights reserved.
# Use is subject to license terms.
#
import sys
from gevent import sleep
from six import reraise
from infi.pyutils.lazy import cached_method
from infi.pyutils.contexts import contextmanager
from infi.pyutils.decorators import _ipython_inspect_module, wraps

from .. import errors
from .utils import SelfLoggerMixin
from .protocol import (RPCCall, encode_rpc_call, decode_rpc_result, RPC_RESULT_CODE_SUCCESS, RPC_RESULT_CODE_ERROR,
                       RPC_RESULT_CODE_DEFERRED)

POLL_SLEEP_INTERVAL_IN_SECONDS = 0.1


class DeferredResult(object):
    """Abstract object for a deferred result."""

    def poll(self):
        """
        Polls for the result.

        :return: `True` if there is a result, `False` if not.
        """
        raise NotImplementedError()

    def is_done(self, poll=False):
        """
        Returns `True` if there is a result, can optionally poll.
        """
        raise NotImplementedError()

    def get_result(self):
        """
        Blocks for a result and returns it or raises an error if the result was an error.
        """
        raise NotImplementedError()


class ImmediateResult(DeferredResult):
    """Deferrable that already contains a result."""

    def __init__(self, success, result):
        super(ImmediateResult, self).__init__()
        self.success = success
        self.result = result

    def poll(self):
        return True

    def is_done(self, poll=False):
        return True

    def get_result(self):
        return self._result()

    def _result(self):
        if self.success:
            return self.result
        else:
            raise errors.RPCException.from_dict(self.result)


class AsyncDeferredResult(DeferredResult):
    """Deferred result that polls for the result."""

    def __init__(self, client, request_id, poll_sleep_interval=POLL_SLEEP_INTERVAL_IN_SECONDS):
        self.client = client
        self.request_id = request_id
        self.poll_sleep_interval = poll_sleep_interval
        self.result = None
        self.exception = None
        self.finished = False
        self.success = None

    def poll(self):
        """
        Polls for the result from the server.

        :return: `True` if there is a result, `False` if not.
        """
        if self.finished:
            return True

        server_response = self.client.query_async_request(self.request_id)
        if server_response['code'] == RPC_RESULT_CODE_DEFERRED:
            return False

        self.finished = True
        self.success = server_response['code'] == RPC_RESULT_CODE_SUCCESS
        if self.success:
            self.result = server_response['result']
        else:
            self.exception = server_response['exception']

        return self.finished

    def is_done(self, poll=False):
        """
        Returns `True` if there is a result, can optionally poll.
        """
        if poll:
            return self.poll()
        else:
            return self.finished

    def get_result(self):
        """
        Blocks for a result and returns it or raises an error if the result was an error.
        """
        if self.finished:
            return self._result()
        while not self.poll():
            sleep(self.poll_sleep_interval)
        return self._result()

    def _result(self):
        if self.success:
            return self.result
        else:
            raise errors.RPCException.from_dict(self.exception)

    def __repr__(self):
        return "{}.{}(request_id={}, client={})".format(self.__class__.__module__, self.__class__.__name__,
                                                        self.request_id, self.client)

    def __str__(self):
        return repr(self)


class Client(SelfLoggerMixin):  # pragma: no cover
    """
    RPC client that supports long asynchronous operations.

    :param transport: transport object to use (e.g. :py:class:`ZeroRPCClientTransport`)
    """
    def __init__(self, transport, poll_sleep_interval=POLL_SLEEP_INTERVAL_IN_SECONDS):
        self._cache = dict()
        self._transport = transport
        self.poll_sleep_interval = poll_sleep_interval

    def call(self, method_name, *args, **kwargs):
        try:
            async = kwargs.pop('async_rpc', False)
            poll_sleep_interval = kwargs.pop('async_rpc_poll_interval', None)

            self._transport.connect()

            rpc_call = RPCCall(method_name, args, kwargs)
            arg = encode_rpc_call(rpc_call)

            self.log_debug("sending RPC request {}".format(rpc_call))
            result = self._transport.call(arg)

            return self._handle_rpc_result(result, rpc_call, async, poll_sleep_interval)
        except errors.TimeoutExpired:
            exc_info = sys.exc_info()
            self.log_debug("got {}: {}".format(exc_info[1].__class__.__name__, exc_info[1]))
            try:
                self._transport.close()
            except:
                close_exc_info = sys.exc_info()
                self.log_debug("encountered {} error on socket close: {}".format(close_exc_info[1].__class__.__name__,
                                                                                 close_exc_info[1]))
            reraise(exc_info[0], exc_info[1], exc_info[2])

    def _handle_rpc_result(self, result_dict, rpc_call, async=False, poll_sleep_interval=None):
        try:
            code, result = decode_rpc_result(result_dict)
        except (errors.InvalidRPCMethod, errors.InvalidCallArguments) as error:
            self.log_error("sent invalid RPC call {}: {}".format(rpc_call, error))
            self._handle_invalid_rpc_exception(error)
            assert False  # should never reach here since _handle_invalid_rpc_exception will raise or abort
            return

        self.log_debug("got RPC response for {}: {}".format(rpc_call, result_dict))
        if code in (RPC_RESULT_CODE_SUCCESS, RPC_RESULT_CODE_ERROR):
            deferred = ImmediateResult(code == RPC_RESULT_CODE_SUCCESS, result)
        else:
            deferred = self._create_deferred_result(result, poll_sleep_interval)

        if async:
            return deferred
        else:
            return deferred.get_result()

    def _create_deferred_result(self, request_id, poll_sleep_interval):
        result = AsyncDeferredResult(self, request_id, self.poll_sleep_interval)
        if poll_sleep_interval is not None:
            result.poll_sleep_interval = poll_sleep_interval
        return result

    def _handle_invalid_rpc_exception(self, error):
        _type, value, traceback = sys.exc_info()
        reraise(_type, value, traceback)

    @cached_method
    def _get_rpc_method_names(self):
        return self.get_rpc_method_names()

    @cached_method
    def _get_rpc_ipython_argspec(self, method_name):
        return self.get_rpc_ipython_argspec(method_name)

    def __getattr__(self, method_name):
        attr = lambda *args, **kwargs: self.call(method_name, *args, **kwargs)
        attr.__name__ = "<bound method {}.{} of {!r}>".format(self.__class__.__name__, method_name, self)
        attr.rpc_method_name = method_name
        attr.rpc_call = True
        return attr

    def __repr__(self):
        return "<infi.rpc:{} 0x{:x}/{}>".format(self.__class__.__name__, id(self), repr(self._transport))

    def __str__(self):
        return repr(self)

    def __dir__(self):
        return dir(super(Client, self)) + self._get_rpc_method_names()


class IPython_Mixin(object):
    """Mixin that shortcircuits IPython methods called when hitting <tab> so they won't get sent to the remote end."""
    def _getAttributeNames(self):  # pragma: no cover
        return dir(self)

    def trait_names(self):  # pragma: no cover
        return []


def strictify(client):
    """A client that dies on invalid RPC calls, useful when building a strict/production system"""
    def _handle_invalid_rpc_exception(error):
        # FIXME
        from izbox.utils import die_now_and_dont_look_back
        reason = "invalid RPC exception: {}".format(error)
        die_now_and_dont_look_back(reason=reason, exc_info=sys.exc_info())
    client._handle_invalid_rpc_exception = _handle_invalid_rpc_exception


def mul_by_two_or_min_one(n):
    """Simple timeout calculation: multiple the max response time by 2 or at least a minimum of 1 second."""
    return max(n * 2, 1)


class AutoTimeoutClient(Client):
    """
    RPC client that supports long asynchronous operations and negotaties the timeout for a request with the server.

    :param transport: transport object to use (e.g. :py:class:`ZeroRPCClientTransport`)
    """

    def __init__(self, transport, poll_sleep_interval=POLL_SLEEP_INTERVAL_IN_SECONDS, timeout_calc_func=mul_by_two_or_min_one):
        super(AutoTimeoutClient, self).__init__(transport, poll_sleep_interval)
        self._timeout_calc_func = timeout_calc_func
        self._server_max_response_time = None
        self._auto_timeout_set = False

    def call(self, method_name, *args, **kwargs):
        if not self._auto_timeout_set:
            max_response_time = self.get_server_max_response_time()
            if max_response_time != -1:
                new_timeout = self._timeout_calc_func(max_response_time)
                self._transport.set_timeout(new_timeout)
                self._auto_timeout_set = True
        return super(AutoTimeoutClient, self).call(method_name, *args, **kwargs)

    def get_server_max_response_time(self):
        if self._server_max_response_time is not None:
            return self._server_max_response_time
        try:
            self._server_max_response_time = super(AutoTimeoutClient, self).call('get_rpc_max_response_time')
        except errors.InvalidRPCMethod:
            self._server_max_response_time = -1
        return self._server_max_response_time


@contextmanager
def patched_ipython_getargspec_context(client):
    original = _ipython_inspect_module.getargspec

    @wraps(original)
    def patched(func):
        if hasattr(func, "rpc_call") and getattr(func, "rpc_call"):
            return client.get_rpc_ipython_argspec(getattr(func, "rpc_method_name", func.__name__))
        return original(func)
    _ipython_inspect_module.getargspec = patched
    _ipython_inspect_module.getargspec = patched
    try:
        yield
    finally:
        _ipython_inspect_module.getargspec = original



#
# Copyright 2013 Infinidat Ltd. All rights reserved.
# Use is subject to license terms.
#

import sys
import functools
import gevent
import gevent.event
from time import time
from inspect import getcallargs
from infi.logging.plugins.request_id_tag import get_tag as get_request_id_tag, request_id_tag_context
from infi.gevent_utils.gevent_loop import GeventLoop

from ..errors import InvalidCallArguments, InvalidAsyncToken
from .protocol import (encode_rpc_result_exc_info, encode_rpc_result_success, encode_rpc_result_deferred,
                       encode_rpc_result_invalid_rpc_method_exc_info, decode_rpc_call)
from .service import is_rpc_deferred_method, is_rpc_immediate_method
from .utils import SelfLoggerMixin


RESULT_GARBAGE_COLLECTOR_INTERVAL = 5
INTERVAL_TO_KEEP_RESULT_IN_SECONDS = 5
MAX_RESPONSE_TIME = 0.5


def validate_call_arguments(func, args, kwargs):
    """
    Makes sure that ``args`` and ``kwargs`` satisfy ``func``'s signature.

    :raises InvalidCallArguments: if ``func``'s signature is not compatible with ``args`` and ``kwargs``
    """
    try:
        getcallargs(func, *args, **kwargs)
    except Exception as error:
        raise InvalidCallArguments(str(error))


class ServerBase(SelfLoggerMixin):  # pragma: no cover
    """
    Base class for an RPC server.

    :param transport: a transport layer object (e.g. :py:class:`ZeroRPCServerTransport` instance)
    :param service: a service object (e.g. a child instance of :py:class:`ServiceBase`)
    """
    def __init__(self, transport, service):
        self._transport = transport
        self._service = service
        self._handler_map = dict()
        self._shutdown_event = gevent.event.Event()

    def _build_handler_map(self):
        method_names = self._service._get_rpc_method_names()
        for name in method_names:
            method = getattr(self._service, name)
            handler = self._map_method_to_handler(name, method)
            self._set_handler_method_map(handler, method, name)

    def _set_handler_method_map(self, handler, method, name=None):
        if name is None:
            name = method.__name__
        self._handler_map[name] = (self._bind_handler_to_method(handler, method), method)

    def _bind_handler_to_method(self, handler, method):
        bound_handler = functools.partial(handler, method)
        functools.update_wrapper(bound_handler, method)
        return bound_handler

    def _map_method_to_handler(self, name, method):
        raise NotImplementedError()

    def bind(self):
        """Start accepting new RPCs. Also clears the `shutdown_event`."""
        self._shutdown_event.clear()
        self._build_handler_map()
        self._transport.bind(self._on_request)

    def unbind(self):
        """Stop accepting new RPCs. Requests that are still in process will not be cancelled."""
        self._transport.unbind()

    def request_shutdown(self):
        """
        Sets the ``_shutdown_event`` event that can also be queried by calling :py:meth:`is_shutdown_requested`.
        By itself it doesn't do anything more but subclasses can wait on the event to implement their own shutdown
        strategy.
        """
        self._shutdown_event.set()

    def is_shutdown_requested(self):
        """
        :return: `True` if a shutdown was requested.
        """
        return self._shutdown_event.is_set()

    def _on_request(self, arg):
        rpc_call = decode_rpc_call(arg)
        with request_id_tag_context(title=None, tag=rpc_call.request_id_tag):
            rpc_call.request_id_tag = get_request_id_tag()  # might get assigned if was None
            self.log_debug("received RPC request {}".format(rpc_call))

            handler, method = self._handler_map.get(rpc_call.method, (None, None))
            if handler is None:
                self.log_error("received invalid RPC method {}, returning error".format(rpc_call.method))
                return encode_rpc_result_invalid_rpc_method_exc_info(rpc_call.method)
            try:
                validate_call_arguments(method, rpc_call.args, rpc_call.kwargs)
            except InvalidCallArguments:
                self.log_error("received invalid RPC arguments for method {}, returning error".format(rpc_call.method))
                return encode_rpc_result_exc_info(sys.exc_info())

            result = handler(rpc_call)
            self.log_debug("replying to RPC request {} with {}".format(rpc_call, result))
            return result

    def _execute_call(self, method, rpc_call):
        with request_id_tag_context(title=None, tag=rpc_call.request_id_tag):
            try:
                result = method(*rpc_call.args, **rpc_call.kwargs)
                return encode_rpc_result_success(result)
            except BaseException:
                exc_info = sys.exc_info()
                self.log_exception("RPC call failed {}".format(rpc_call))
                return encode_rpc_result_exc_info(exc_info)


class SyncServer(ServerBase):
    """
    A server that executes every RPC synchronously, which means that the server will not return a response to the client
    until the request completes.

    Although it's very simple, for most implementations you probably want to use :py:class:`Server` instead since with
    this server the client can't tell if the server is stuck or the request is still in progress.

    :param transport: a transport layer object (e.g. :py:class:`ZeroRPCServerTransport` instance)
    :param service: a service object (e.g. a child instance of :py:class:`ServiceBase`)
    """
    def _map_method_to_handler(self, name, method):
        return self._execute_call


class AsyncRequest(object):
    QUEUED = 1
    PROCESSING = 2
    RESULT_AVAILABLE = 3

    def __init__(self, rpc_call):
        self.state = AsyncRequest.QUEUED
        self.rpc_call = rpc_call
        self.greenlet = None

    def spawn(self, method, args, kwargs):
        self.state = AsyncRequest.PROCESSING
        self.greenlet = gevent.spawn(method, *args, **kwargs)
        self.greenlet.link(self._result_available)

    def has_result(self):
        return self.state == AsyncRequest.RESULT_AVAILABLE

    def get_result(self):
        return self.greenlet.get(block=False)

    def wait(self, timeout):
        self.greenlet.join(timeout)

    def _result_available(self, _):
        self.state = AsyncRequest.RESULT_AVAILABLE
        self.finish_time = time()


class AsyncMixin(object):
    def __init__(self, result_garbage_collector_interval=RESULT_GARBAGE_COLLECTOR_INTERVAL,
                 interval_to_keep_result=INTERVAL_TO_KEEP_RESULT_IN_SECONDS):
        self._deferred_requests = dict()
        self._interval_to_keep_result = interval_to_keep_result
        self._result_garbage_collector = GeventLoop(result_garbage_collector_interval,
                                                    self.garbage_collect_deferred_results)

    def _find_or_spawn_async_request(self, method, call):
        async_request = self._deferred_requests.get(call.uuid, None)
        if async_request is None:
            async_request = AsyncRequest(call)
            self._deferred_requests[call.uuid] = async_request
            async_request.spawn(self._execute_call, [method, call], dict())
        return async_request

    def _async_handler(self, method, call):
        async_request = self._find_or_spawn_async_request(method, call)
        return self._build_async_response(async_request)

    def _query_async_request(self, uuid):
        async_request = self._deferred_requests.get(uuid, None)
        if async_request is None:
            raise InvalidAsyncToken(uuid)
        return self._build_async_response(async_request)

    def _set_query_async_request_to_handler_map(self):
        self._set_handler_method_map(self._execute_call, self._query_async_request, 'query_async_request')

    def _build_async_response(self, request):
        if request.has_result():
            del self._deferred_requests[request.rpc_call.uuid]
            try:
                return request.get_result()
            except BaseException:
                exc_info = sys.exc_info()
                return encode_rpc_result_exc_info(exc_info)
        else:
            return encode_rpc_result_deferred(request.rpc_call.uuid)

    def garbage_collect_deferred_results(self):
        """Removes unclaimed results that finished more than ``interval_to_keep_result`` seconds ago."""
        uuids_to_delete = set()
        now = time()
        for uuid, req in self._deferred_requests.items():
            if req.state == AsyncRequest.RESULT_AVAILABLE and now > (req.finish_time + self._interval_to_keep_result):
                uuids_to_delete.add(uuid)
        for uuid in uuids_to_delete:
            del self._deferred_requests[uuid]

    def start_garbage_collect_deferred_results(self):
        """Starts the unclaimed results garbage collection greenlet"""
        self._result_garbage_collector.start()

    def stop_garbage_collect_deferred_results(self):
        """Stops the unclaimed results garbage collection greenlet"""
        self._result_garbage_collector.stop()

    def deferred_requests_count(self):
        """:return: count of deferred requests (in-progress and finished requests yet to be claimed by clients)"""
        return len(self._deferred_requests)

    def has_deferred_requests(self):
        """
        :return: ``True`` if there are active deferred requests (either in-progress or waiting a client to poll their
                 result)
        """
        return bool(self._deferred_requests)

    def wait_for_empty_deferred_requests(self, timeout=None, poll_interval=0.1):
        """
        Waits for all deferred requests to complete and either get polled or expire.

        :param timeout: seconds to wait before giving up, or ``None`` to block forever
        :param poll_interval: how often to poll and check if all requests are done
        :return: ``True`` if all deferred requests completed and got polled or expired, ``False`` if there are still
                 active deferred requests.
        """
        if timeout is None:
            timeout = 86400 * 365  # 1 year - no one waits one year for a method.
        start_time = time()
        end_time = start_time + timeout
        while self.has_deferred_requests() and time() < end_time:
            gevent.sleep(poll_interval)
            self.garbage_collect_deferred_results()  # manually try to do garbage collection to speed things up.
        return not self.has_deferred_requests()


class AsyncServer(ServerBase, AsyncMixin):
    """
    A server that executes every RPC asynchronously, which means for every RPC it will immediately return a deferred
    result with a token so the client can use to poll for the actual result.

    Since this server *always* return a deferred result this may cause unnecessary round-trips if the RPC never blocks
    (for example, a `ping` call).

    For most implementations, you probably want to use :py:class:`Server` instead.


    :param transport: a transport layer object (e.g. :py:class:`ZeroRPCServerTransport` instance)
    :param service: a service object (e.g. a child instance of :py:class:`ServiceBase`)
    :param result_garbage_collector_interval: how often (in seconds) should a garbage collector run and remove
                                              unclaimed results (response for a request that the client didn't
                                              poll to get)
    :param interval_to_keep_result: how much time (in seconds) should keep an unclaimed result. Since the clients
                                    poll to get the result, the interval should be greater than the client's poll
                                    interval (usually twice the time of the client's poll interval).
    """
    def __init__(self, transport, service, result_garbage_collector_interval=RESULT_GARBAGE_COLLECTOR_INTERVAL,
                 interval_to_keep_result=INTERVAL_TO_KEEP_RESULT_IN_SECONDS):
        ServerBase.__init__(self, transport, service)
        AsyncMixin.__init__(self, result_garbage_collector_interval, interval_to_keep_result)
        self._set_query_async_request_to_handler_map()

    def _map_method_to_handler(self, name, method):
        if is_rpc_immediate_method(method):
            return self._execute_call
        return self._async_handler

    # FIXME do we want to provide this here on bind? should we offer new "start/stop" methods that combine these?
    def bind(self):
        """
        Start accepting new RPCs and the deferred results garbage collector greenlet.
        Also clears the `shutdown_event`.
        """
        super(AsyncServer, self).bind()
        self.start_garbage_collect_deferred_results()

    def unbind(self):
        """
        Stop accepting new RPCs and the deferred results garbage collector greenlet.
        Requests that are still in process will not be cancelled.
        """
        try:
            super(AsyncServer, self).unbind()
        finally:
            self.stop_garbage_collect_deferred_results()


class Server(AsyncServer):
    """
    A server that can handle synchronous RPCs and asynchronous RPCs and guarantee a maximum time before replying to
    a client.

    :param transport: a transport layer object (e.g. :py:class:`ZeroRPCServerTransport` instance)
    :param service: a service object (e.g. a child instance of :py:class:`ServiceBase`)
    :param result_garbage_collector_interval: how often (in seconds) should a garbage collector run and remove
                                              unclaimed results (response for a request that the client didn't
                                              poll to get)
    :param interval_to_keep_result: how much time (in seconds) should keep an unclaimed result. Since the clients
                                    poll to get the result, the interval should be greater than the client's poll
                                    interval (usually twice the time of the client's poll interval).
    :param max_response_time: how much time to block and wait for the service's method to return before returning
                              a deferred response.
    """
    def __init__(self, transport, service,
                 result_garbage_collector_interval=RESULT_GARBAGE_COLLECTOR_INTERVAL,
                 interval_to_keep_result=INTERVAL_TO_KEEP_RESULT_IN_SECONDS,
                 max_response_time=MAX_RESPONSE_TIME):
        AsyncServer.__init__(self, transport, service, result_garbage_collector_interval, interval_to_keep_result)
        self._max_response_time = max_response_time
        self._set_handler_method_map(self._execute_call, self.get_rpc_max_response_time, 'get_rpc_max_response_time')

    def _map_method_to_handler(self, name, method):
        if is_rpc_deferred_method(method):
            return self._async_handler

        if self._max_response_time is None or is_rpc_immediate_method(method):
            return self._execute_call
        else:
            return self._semi_async_handler

    def _semi_async_handler(self, method, call):
        async_request = self._find_or_spawn_async_request(method, call)
        if async_request.state == AsyncRequest.RESULT_AVAILABLE:
            return self._build_async_response(async_request)
        async_request.wait(self._max_response_time)
        return self._build_async_response(async_request)

    def get_rpc_max_response_time(self):
        return self._max_response_time



from .base import (ServiceBase, rpc_call, rpc_deferred, rpc_immediate, synchronized,
                   ServiceWithSynchronized, SyncServer, AsyncServer, Server, Client,
                   AutoTimeoutClient, IPython_Mixin, DeferredResult, ImmediateResult, AsyncDeferredResult, strictify,
                   patched_ipython_getargspec_context)
from .zerorpc import ZeroRPCClientTransport, ZeroRPCServerTransport

__all__ = [
    'ServiceBase', 'rpc_call', 'rpc_deferred', 'rpc_immediate', 'synchronized',
    'ServiceWithSynchronized', 'SyncServer', 'AsyncServer', 'Server', 'Client',
    'AutoTimeoutClient',
    'IPython_Mixin', 'DeferredResult', 'ImmediateResult', 'AsyncDeferredResult',
    'strictify',
    'ZeroRPCClientTransport', 'ZeroRPCServerTransport',
    'patched_ipython_getargspec_context'
]



#
# Copyright 2013 Infinidat Ltd. All rights reserved.
# Use is subject to license terms.
#
from .service import (ServiceBase, rpc_call, rpc_deferred, rpc_immediate, synchronized, SynchronizedMixin,
                      ServiceWithSynchronized)
from .server import ServerBase, SyncServer, AsyncServer, Server
from .client import (Client, AutoTimeoutClient, IPython_Mixin, DeferredResult, ImmediateResult, AsyncDeferredResult,
                     strictify, patched_ipython_getargspec_context)
from .utils import format_request, SelfLoggerMixin
from .protocol import (ServerTransport, ClientTransport, encode_rpc_result_exc_info, encode_rpc_result_exception,
                       encode_rpc_result_invalid_rpc_method_exc_info)


__all__ = [
    'ServiceBase', 'rpc_call', 'rpc_deferred', 'rpc_immediate', 'synchronized', 'SynchronizedMixin',
    'ServiceWithSynchronized',
    'ServerBase', 'SyncServer', 'AsyncServer', 'Server',
    'Client', 'AutoTimeoutClient', 'IPython_Mixin', 'DeferredResult', 'ImmediateResult', 'AsyncDeferredResult',
    'strictify', 'format_request', 'SelfLoggerMixin',
    'ServerTransport', 'ClientTransport', 'encode_rpc_result_exc_info',
    'encode_rpc_result_exception', 'encode_rpc_result_invalid_rpc_method_exc_info',
    'patched_ipython_getargspec_context'
]



#
# Copyright 2013 Infinidat Ltd. All rights reserved.
# Use is subject to license terms.
#

import types
from gevent.lock import RLock
from infi.pyutils.decorators import wraps

from ..errors import InvalidRPCMethod
from .utils import SelfLoggerMixin


def rpc_call(func):
    """
    Decorator wrapping an RPC method that marks it as such (an RPC method). Use this decorator to mark which methods
    in a service should be exposed as RPC methods.
    """
    @wraps(func)
    def decorator(*args, **kwargs):
        return func(*args, **kwargs)
    decorator.rpc_call = True
    return decorator


def rpc_deferred(func):
    """
    Decorator wrapping an RPC method that marks it as a method that will most likely block and thus the server should
    immediately return a deferred result and not wait.
    """
    decorator = rpc_call(func)
    decorator.rpc_deferred = True
    return decorator


def rpc_immediate(func):
    """Decorator wrapping an RPC method that marks it as a method that will always return a result without blocking."""
    decorator = rpc_call(func)
    decorator.rpc_immediate = True
    return decorator


def is_rpc_deferred_method(func):
    return getattr(func, 'rpc_deferred', False)


def is_rpc_immediate_method(func):
    return getattr(func, 'rpc_immediate', False)


def _is_rpc_call_function(func):
    return getattr(func, 'rpc_call', False)


def _is_rpc_call_method(method):
    return isinstance(method, types.MethodType) and _is_rpc_call_function(method.__func__)


class ServiceBase(SelfLoggerMixin):
    """
    Base class for an RPC service. A service is a logical unit that exposes an RPC API.

    To create a service, you need to inherit from this class or other classes such as
    :py:class:`ServiceWithSynchronized` and decorate the methods you want to expose.

    For example::

        class MyService(infi.rpc.ServiceBase):
            @rpc_call
            def my_rpc_exposed_method(self, my_arg):
                return 42

    To expose the service as a server, see the :ref:`Server` class for example.
    """

    @rpc_call
    def get_rpc_method_names(self):
        """
        Provides introspection information to the client side.

        :return: list of callable RPC method names
        :rtype: list of strings
        """
        return self._get_rpc_method_names()

    @rpc_call
    def get_rpc_ipython_argspec(self, method_name):
        """
        Provides introspection information about an RPC method to clients that use IPython.

        :param method_name: method name string
        :return: IPython's getargspec result
        :raises InvalidRPCMethod: if the method is not an RPC method
        """
        from infi.pyutils.decorators import _ipython_inspect_module
        try:
            method = getattr(self, method_name)
        except AttributeError:
            raise InvalidRPCMethod(method_name)

        if not _is_rpc_call_method(method):
            raise InvalidRPCMethod(method_name)

        return _ipython_inspect_module.getargspec(method)

    @rpc_call
    def ping(self):
        """RPC method to just ping the service."""
        pass

    def _get_rpc_method_names(self):
        """
        Override this method to provide a custom list of RPC-exposed method names.

        :return: list of exposed RPC method names
        :rtype: list of strings
        """
        return [name for name in dir(self) if _is_rpc_call_method(getattr(self, name))]


class SynchronizedMixin(object):
    def __init__(self):
        self.synchronized_mutex = RLock()


def synchronized(func):
    """
    Decorator that synchronizes execution of all the `synchronized` decorated methods.
    This is useful if there are methods in the service that should not execute in parallel.

    .. note::
        To support synchronized methods, inherit from :py:class:`ServiceWithSynchronized` instead of
        :py:class:`ServiceBase`.
    """
    @wraps(func)
    def decorator(*args, **kwargs):
        if not args or not isinstance(args[0], SynchronizedMixin):
            raise ValueError("method {} if not part of a Synchronized service".format(func))
        with args[0].synchronized_mutex:
            return func(*args, **kwargs)
    return decorator


class ServiceWithSynchronized(ServiceBase, SynchronizedMixin):
    """
    Adds support for synchronizing RPC method calls.

    When you need to ensure that some of the RPC methods will not execute in parallel (e.g. synchronize file writes) use
    this as a base class and decorate the methods with the :py:func:`synchronized`.

    For example::

        class FileLogger(infi.rpc.ServiceWithSynchronized):
            @rpc_call
            @synchronized
            def write_log(self, message):
                ...

            @rpc_call
            @synchronized
            def rotate(self):
                ...
    """
    def __init__(self, *args, **kwargs):
        ServiceBase.__init__(self)
        SynchronizedMixin.__init__(self)



#
# Copyright 2013 Infinidat Ltd. All rights reserved.
# Use is subject to license terms.
#

import zmq
import zerorpc
import logbook
from infi.logging.plugins.request_id_tag import request_id_tag
from infi.gevent_utils.safe_greenlets import safe_spawn
from gevent import sleep

from .. import base

# FIXME
from gevent import spawn as spawn_silent


ANYHOST = "0.0.0.0"

logger = logbook.Logger("infi.rpc")


class _ZeroRPCServer(zerorpc.Server, base.SelfLoggerMixin):
    def __init__(self, address, callback, name=None, context=None, pool_size=None):
        self._address = address
        zerorpc.Server.__init__(self, dict(on_call=callback), name, context, pool_size)

        zerorpc.Server.bind(self, self._address)

        # when we go down, we want to drop all the packets in the air
        self._events.setsockopt(zmq.LINGER, 0)

        # zerorpc.Server.run blocks, so we do what it does except for the blocking part
        self._acceptor_task = safe_spawn(self._acceptor)

    @request_id_tag(tag="acceptor")
    def _acceptor(self):
        while True:
            initial_event = self._multiplexer.recv()
            self._task_pool.add(spawn_silent(self._async_task, initial_event))

    # FIXME
    # @die_on_greenlet_exceptions
    def _async_task(self, initial_event):
        # We had to replace zerorpc's _async_task implementation due to two reasons:
        # 1. Original implementation doesn't handle exceptions like we want, specifically in its response-builder
        #    code.
        # 2. We don't want heartbeats because zerorpc's heartbeat implementation may cause active requests in the server
        #    to get killed (they are greenlets) in middle of processing if the heartbeat detects a lost peer.
        channel = self._multiplexer.channel(initial_event)
        bufchan = zerorpc.BufferedChannel(channel)
        event = bufchan.recv()
        try:
            self._context.hook_load_task_context(event.header)
            functor = self._methods.get(event.name, None)
            if functor is not None:
                functor.pattern.process_call(self._context, bufchan, event, functor)
            else:
                self.log_debug("invalid RPC method received: {}".format(event.name))
                result = base.encode_rpc_result_invalid_rpc_method_exc_info(event.name)

                if hasattr(bufchan, 'new_event'):
                    reply_event = bufchan.new_event('OK', (result,), self._context.hook_get_task_context())
                else:
                    reply_event = bufchan.create_event('OK', (result,), self._context.hook_get_task_context())
                self._context.hook_server_after_exec(event, reply_event)
                bufchan.emit_event(reply_event)
        except:
            self.log_exception("unhandled exception in zerorpc transport while handling event {}".format(event))
            raise
        finally:
            bufchan.close()

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return "<{}.{}({})>".format(self.__class__.__module__, self.__class__.__name__,
                                    self._address if self._address else "unbound")


class ZeroRPCServerTransport(base.ServerTransport):
    """
    Transport layer implementation using ZeroRPC. We delegate almost everything to our ``_ZeroRPCServer``
    implementation.
    """

    @classmethod
    def create_tcp(cls, port, host=ANYHOST, *args, **kwargs):
        # One would say that 0.0.0.0 is too open but we'd like to be able to do RPC from outside the system for tests.
        # In the field we will have ipfilter blocking connections from outside the system.
        return ZeroRPCServerTransport("tcp://{}:{}".format(host, port), *args, **kwargs)

    def __init__(self, address, name=None, context=None, pool_size=None):
        self.address = address
        self.name = name
        self.context = context
        self.pool_size = pool_size
        self.zerorpc = None

    def bind(self, callback):
        self.zerorpc = _ZeroRPCServer(self.address, callback, self.name, self.context, self.pool_size)

    def unbind(self):
        if self.zerorpc:
            self.zerorpc.close()
            self._zmq_error_workaround()
            self.zerorpc = None

    def _zmq_error_workaround(self):
        # See https://github.com/zeromq/pyzmq/issues/129
        # Traceback (most recent call last):
        #   File "/Users/guy/gitserver/izbox/src/izbox/rpc/tests.py", line 24, in test_up_down_loop
        #     server.setup()
        #   File "/Users/guy/gitserver/izbox/src/izbox/rpc/zerorpc.py", line 23, in setup
        #     super(Server, self).bind("tcp://127.0.0.1:{}".format(self._port))
        #   File ".../site-packages/zerorpc-0.4.1-py2.7.egg/zerorpc/socket.py", line 43, in bind
        #     return self._events.bind(endpoint, resolve)
        #   File ".../site-packages/zerorpc-0.4.1-py2.7.egg/zerorpc/events.py", line 218, in bind
        #     r.append(self._socket.bind(endpoint_))
        #   File "socket.pyx", line 465, in zmq.core.socket.Socket.bind (zmq/core/socket.c:4749)
        # ZMQError: Address already in use
        sleep(0.05)



from logbook import Logger

logger = Logger('infi.rpc')


class SelfLoggerMixin(object):
    def log_debug(self, format, *args, **kwargs):
        self.get_logger().debug("{} {}".format(self, format), *args, **kwargs)

    def log_error(self, format, *args, **kwargs):
        self.get_logger().error("{} {}".format(self, format), *args, **kwargs)

    def log_exception(self, format, *args, **kwargs):
        self.get_logger().exception("{} {}".format(self, format), *args, **kwargs)

    def get_logger(self):
        return logger


def format_request(method, args, kwargs):
    return "{}(*{}, **{})".format(method, args, kwargs)



import jsonpickle
from sys import exc_info
from .json_traceback import format_tb


class RPCException(Exception):
    should_be_logged = True
    log_with_traceback = True

    def to_dict(self, with_traceback=None):
        return dict(pickled=jsonpickle.encode(self), message=str(self), args=self.args)

    @classmethod
    def from_dict(cls, result_dict):
        error = jsonpickle.decode(result_dict['pickled'])
        for key, value in result_dict.items():
            if key == 'pickled':
                continue
            setattr(error, key, value)
        return error


class RPCUserException(RPCException):
    pass


class RPCFrameworkException(RPCException):
    pass


class RPCTransportException(RPCFrameworkException):
    pass


class TimeoutExpired(RPCTransportException):
    pass


class InternalServerError(RPCFrameworkException):
    def __init__(self, original_exception):
        super(InternalServerError, self).__init__("type={!r}, value={!r}".format(type(original_exception), original_exception))


class InvalidReturnValue(RPCFrameworkException):
    pass


class InvalidRPCMethod(RPCFrameworkException):
    should_be_logged = False


class InvalidCallArguments(RPCFrameworkException):
    should_be_logged = False


class InvalidAsyncToken(RPCFrameworkException):
    should_be_logged = False


# FIXME not sure we need the "NotSetup" errors anymore...
class ClientNotSetupError(RPCFrameworkException):
    def __init__(self):
        super(ClientNotSetupError, self).__init__("client.setup() was not called or closed and not setup again")


class ServerNotSetupError(RPCFrameworkException):
    def __init__(self):
        super(ServerNotSetupError, self).__init__("server.setup() was not called or closed and not setup again")


class RequestAbortedError(RPCFrameworkException):
    def __init__(self):
        super(RequestAbortedError, self).__init__("request aborted while being processed")



#
# Copyright 2013 Infinidat Ltd. All rights reserved.
# Use is subject to license terms.
#
import sys
from munch import munchify
from uuid import uuid4
from infi.logging.plugins.request_id_tag import get_tag as get_request_id_tag

from .. import json_traceback
from ..errors import RPCException, InvalidReturnValue, InvalidRPCMethod, InternalServerError
from .utils import format_request

RPC_RESULT_CODE_SUCCESS = 'success'
RPC_RESULT_CODE_DEFERRED = 'deferred'
RPC_RESULT_CODE_ERROR = 'error'
RPC_RESULT_CODES = (RPC_RESULT_CODE_SUCCESS, RPC_RESULT_CODE_DEFERRED, RPC_RESULT_CODE_ERROR)


class ServerTransport(object):
    """
    This is a bridge between the transport layer (say ZeroRPC) and the high level object layer on the server side.
    It basically provides two methods to start accepting requests and stop accepting requests: ``bind`` and ``unbind``.
    """
    def bind(self, callback):
        """
        Start accepting requests from the transport layer. ``callback`` is a method that receives a single dict argument
        and will decode our RPC layer and then encode the result using ``protocol``.
        """
        raise NotImplementedError()

    def unbind(self):
        """
        Stop accepting requests from the transport layer.
        """
        raise NotImplementedError()


class ClientTransport(object):
    """
    This is a bridge between the transport layer (say ZeroRPC) and the high level object layer on the client side.
    """
    def connect(self):
        raise NotImplementedError()

    def close(self):
        raise NotImplementedError()

    def call(self, arg):
        raise NotImplementedError()

    def set_timeout(self, timeout):
        raise NotImplementedError()


class RPCCall(object):
    def __init__(self, method, args, kwargs, uuid=None, request_id_tag=None):
        self.method = method
        self.args = args
        self.kwargs = kwargs
        self.uuid = uuid
        self.request_id_tag = request_id_tag
        self.formatted_request = format_request(method, args, kwargs)

    def __repr__(self):
        return "(uuid={}, request_id_tag={}) {}".format(self.uuid, self.request_id_tag, self.formatted_request)

    def __str__(self):
        return repr(self)


def encode_rpc_call(rpc_call):
    rpc_call.uuid = str(uuid4()) if rpc_call.uuid is None else rpc_call.uuid
    rpc_call.request_id_tag = get_request_id_tag() if rpc_call.request_id_tag is None else rpc_call.request_id_tag
    meta = dict(method=rpc_call.method, uuid=rpc_call.uuid, request_id_tag=rpc_call.request_id_tag)
    return [meta, rpc_call.args, rpc_call.kwargs]


def decode_rpc_call(arg):
    meta, args, kwargs = arg
    name = meta['method']
    uuid = meta.pop('uuid', None)
    request_id_tag = meta.pop('request_id_tag', None)
    return RPCCall(name, args, kwargs, uuid, request_id_tag)


def encode_rpc_result_success(result):
    return dict(code=RPC_RESULT_CODE_SUCCESS, result=result)


def encode_rpc_result_exc_info(exc_info, with_traceback=None):
    if isinstance(exc_info[1], RPCException):
        result = encode_rpc_result_exception(exc_info[1])
    else:
        result = encode_rpc_result_exception(InternalServerError(exc_info[1]))
    if with_traceback is True or (with_traceback is None and getattr(exc_info[1], 'log_with_traceback', False)):
        result['server_side_traceback'] = json_traceback.format_tb(exc_info[2])
    return result


def encode_rpc_result_invalid_rpc_method_exc_info(name):
    try:
        raise InvalidRPCMethod(name)
    except:
        exc_info = sys.exc_info()
    return encode_rpc_result_exc_info(exc_info)


def encode_rpc_result_exception(exception):
    assert isinstance(exception, RPCException)
    return dict(code=RPC_RESULT_CODE_ERROR, exception=exception.to_dict())

def encode_rpc_result_deferred(uuid):
    return dict(code=RPC_RESULT_CODE_DEFERRED, result=uuid)


def decode_rpc_result(result_dict):
    if not isinstance(result_dict, dict):
        raise InvalidReturnValue(result_dict)
    result = munchify(result_dict)  # for convenience
    code = result.get('code')
    if code not in RPC_RESULT_CODES:
        raise InvalidReturnValue(result)
    if code in (RPC_RESULT_CODE_DEFERRED, RPC_RESULT_CODE_SUCCESS):
        return code, result.result
    else:
        return code, result.exception



#
# Copyright 2013 Infinidat Ltd. All rights reserved.
# Use is subject to license terms.
#

import sys
import zerorpc
import zmq
import gevent
from zerorpc.channel import BufferedChannel

from .. import base
from .. import errors

LOCALHOST = "127.0.0.1"
CLIENT_TIMEOUT_IN_SECONDS = 10.0


class _ZeroRPCClient(zerorpc.Client):
    """ZeroRPC's client with some changes to remove heartbeat implementation"""

    def __call__(self, method, *args, **kargs):
        # We modify zerorpc's __call__ method to remove the heartbeat feature since it may cause server-side requests
        # to get interrupted during processing if it detects a lost remote and it may also fail if the client and
        # server don't have the same heartbeat frequency set.
        #
        # The following code is taken from zerorpc/core.py:
        timeout = kargs.get('timeout', self._timeout)
        channel = self._multiplexer.channel()
        bufchan = BufferedChannel(channel, inqueue_size=kargs.get('slots', 100))

        xheader = self._context.hook_get_task_context()
        if hasattr(bufchan, 'new_event'):
            request_event = bufchan.new_event(method, args, xheader)
        else:
            request_event = bufchan.create_event(method, args, xheader)
        self._context.hook_client_before_request(request_event)
        bufchan.emit_event(request_event)

        try:
            if kargs.get('async', False) is False:
                return self._process_response(request_event, bufchan, timeout)

            async_result = gevent.event.AsyncResult()
            gevent.spawn(self._process_response, request_event, bufchan, timeout).link(async_result)
            return async_result
        except:
            # XXX: This is going to be closed twice if async is false and
            # _process_response raises an exception. I wonder if the above
            # async branch can raise an exception too, if no we can just remove
            # this code.
            bufchan.close()
            raise


class ZeroRPCClientTransport(base.ClientTransport, base.SelfLoggerMixin):
    """
    Transport layer implementation using ZeroRPC.

    Since the protocol layer handles all the method and argument encapsulation we don't use ZeroRPC's method calling,
    instead we assume (from the client-side) there's a method called ``on_call`` on the server-side implementation and
    we call it with the encoded method call.
    """

    @classmethod
    def create_tcp(cls, port, host=LOCALHOST, *args, **kwargs):
        """Convenience method to create a client over TCP."""
        return ZeroRPCClientTransport("tcp://{}:{}".format(host, port), *args, **kwargs)

    def __init__(self, address, timeout=CLIENT_TIMEOUT_IN_SECONDS):
        self._address = address
        self._timeout = timeout
        self._zerorpc = None

    def connect(self):
        """
        If not already connected, creates a ZeroRPC client to the address specified in the constructor.
        The method is called by the higher layer before every RPC call.
        """
        if self._zerorpc:
            return
        try:
            self._zerorpc = _ZeroRPCClient(connect_to=self._address, timeout=self._timeout)
            self._zerorpc._events.setsockopt(zmq.LINGER, 0)  # when we teardown, we want to discard all messages
        except:
            self._zerorpc = None
            raise

    def close(self):
        """
        If a client exists, closes it and removes it.
        This method may be called when an error occured.
        """
        if not self._zerorpc:
            return
        try:
            self._zerorpc.close()
        finally:
            self._zerorpc = None

    def call(self, arg):
        """
        Calls the server-side transport by assuming an ``on_call`` method on the server-side.
        Translates ZeroRPC's timeout exception to ours.
        """
        try:
            return self._zerorpc.__call__('on_call', arg)
        except zerorpc.exceptions.TimeoutExpired as error:
            self.log_debug("got zerorpc.exceptions.TimeoutExpired: {}".format(error))
            try:
                self.close()
            except:
                exc_info = sys.exc_info()
                self.log_debug("encountered error on socket close: {}".format(exc_info[1]))
            raise errors.TimeoutExpired(error)

    def set_timeout(self, timeout):
        """Sets the timeout to be used from here on when making (new) RPC calls."""
        if self._timeout != timeout:
            self._timeout = timeout
            if self._zerorpc:
                self.close()
                self.connect()

    def __repr__(self):
        return "<zerorpc {}[{}]>".format(self._address, "connected" if self._zerorpc else "not connected")

    def __str__(self):
        return repr(self)



import sys
import linecache
from os import path, pardir, sep

# FIXME
BASEDIR = path.abspath(path.join(path.dirname(__file__), pardir, pardir, pardir))

from infi.gevent_utils.json_utils import encode, decode

def format_tb(tb, limit=None):
    """A shorthand for 'format_list(extract_stack(f, limit))."""
    return format_list(extract_tb(tb, limit))


def extract_tb(tb, limit=None):
    if limit is None:
        if hasattr(sys, 'tracebacklimit'):
            limit = sys.tracebacklimit
    list = []
    n = 0
    while tb is not None and (limit is None or n < limit):
        f = tb.tb_frame
        lineno = tb.tb_lineno
        co = f.f_code
        filename = co.co_filename
        name = co.co_name
        _locals = tb. tb_frame.f_locals
        linecache.checkcache(filename)
        line = linecache.getline(filename, lineno, f.f_globals)
        if line:
            line = line.strip()
        else:
            line = None
        list.append((filename, lineno, name, line, _locals))
        tb = tb.tb_next
        n = n + 1
    return list


def safe_json_repr(value, length_limit=1024):
    try:
        encoded_string = encode(value)
        if length_limit and len(encoded_string) > length_limit:
            raise AssertionError()
        return decode(encoded_string)
    except:
        pass
    try:
        encoded_string = repr(value)
        if length_limit and len(encoded_string) > length_limit:
            raise AssertionError()
        return encoded_string
    except:
        pass
    return object.__repr__(value)


def format_list(extracted_list):
    list = []
    for filename, lineno, name, line, _locals in extracted_list:
        relpath = "{}:{}".format(filename.replace(BASEDIR + sep, ''), str(lineno))
        item = dict(line=line, name=name, locals=dict(), relpath=relpath)
        if isinstance(_locals, dict):
            item['locals'] = {key: safe_json_repr(value) for key, value in _locals.items()}
        list.append(item)
    return list
